#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/power.h>

#include "LUFA/Version.h"
#include "LUFA/Scheduler/Scheduler.h"                // Simple scheduler for task management

#include "usb_midi_descriptors.h"
#include "midi.h"
#include "keypad.h"

volatile uint16_t keypad_state = 0x0000;
volatile uint16_t last_keypad_state = 0x0000;
volatile bool keys_changed = false;

volatile int encoder_value = 0;
volatile uint8_t encoder_state = 0xFF;
volatile uint8_t last_encoder_state = 0xFF;
volatile bool encoder_changed = false;

volatile uint8_t yonk = 0;

ISR(TIMER0_OVF_vect) {
	//Just scan hardware state and store it away.  Take no usb action in this timer interrupt handler.
	scan_keypad();
	scan_encoder();
}

int main(void) {
	
	/* Disable watchdog if enabled by bootloader/fuses */
	MCUSR &= ~(1 << WDRF);
	wdt_disable();

	/* Disable clock division */
	clock_prescale_set(clock_div_1);
	
	// Enable pull-up bus power by turning off PUD (pull-up disable).
	MCUCR &= ~_BV(PUD);

	init_keypad();
	init_timer0();
	init_display();
	init_encoder();
	sei();			//enable global interrupts
		
	/* Initialize Scheduler so that it can be used */
	Scheduler_Init();

	/* Initialize USB Subsystem */
	USB_Init();

	/* Scheduling - routine never returns, so put this last in the main function */
	Scheduler_Start();	//last thing to do!
}

void init_display(void){
	DIGIT1_DDR = 0xFF;			//All outputs
	DIGIT1_PORT = 0xFF;			//Drive active high to turn off
	DIGIT2_DDR = 0xFF; 			//All outputs
	DIGIT2_PORT = 0xFF;			//All high
}

void init_timer0(void){
	TCCR0B = 0b00000101;	//256 prescaler
	//TCCR0B = (_BV(CS00)); //Select clock source internal with no prescaling
	TIMSK0 |= _BV(TOIE0);	//enable overflow interrupt
}

// 4 rows are inputs
// 3 cols are outputs
void init_keypad(){
	KEYPAD_DDR = 0b00000000;		//all inputs to start with -- outputs are enabled on-demand

	//Enable pullups on all row pins (inputs)
	KEYPAD_PORT = _BV(KP_ROW1_PIN) | _BV(KP_ROW2_PIN) | _BV(KP_ROW3_PIN) | _BV(KP_ROW4_PIN);

	//Tri-state the outputs
	KEYPAD_PORT &= ~(_BV(KP_COL1_PIN) | _BV(KP_COL2_PIN) | _BV(KP_COL3_PIN));
}

void init_encoder(){
	ENCODER_DDR = 0b00000000;
	ENCODER_PORT = _BV(ENCODER_LEFT_PIN) | _BV(ENCODER_RIGHT_PIN);	//enable pullups
}

void scan_encoder() {
	if(encoder_changed){
		return;			//still waiting for usb handler to update us
	}
	last_encoder_state = encoder_state;
	encoder_state = (ENCODER_PINS & ENCODER_MASK);

	if(encoder_state != last_encoder_state){
		encoder_changed = true;
	}
}

void scan_keypad() {

	if(keys_changed){		//still waiting to send last change, don't rescan
		return;
	}

	last_keypad_state = keypad_state;
	
	for(uint8_t col = 0; col < 3; col++){
	
		KEYPAD_DDR |= _BV(KP_COL1_PIN + col);  // get the col pin ready for output
		KEYPAD_PORT |= _BV(KP_COL1_PIN + col); // then turn it on
		uint8_t pins = (KEYPAD_PINS & KEYPAD_PIN_MASK);	//read input pins
		KEYPAD_PORT &= ~(_BV(KP_COL1_PIN + col)); // now turn it back off
		KEYPAD_DDR &= ~(_BV(KP_COL1_PIN + col));  // and make input (so Hi-z)

		for(uint8_t row = 0; row < 4; row++){
			uint8_t bit = (3*(3-row)) + col;
			if((pins & (1 << row)) == 0){	//pin pulled low
				keypad_state |= (1 << bit);
			}
			else{
				keypad_state &= ~(1 << bit);	//clear out our slot
			}
		}
	}

	if(keypad_state != last_keypad_state){
		keys_changed = true;
	}
}

TASK(USB_MIDI_Task) {

	/* Select the MIDI IN stream */
	// Believed to be IN as seen by the host.
	Endpoint_SelectEndpoint(MIDI_STREAM_IN_EPNUM);

	if(keys_changed){
		DIGIT2_PORT ^= _BV(DIGIT_DP_PIN);	//toggle 2nd digit decimal point
		send_keypad_data();
		keys_changed = false;
	}

	if(encoder_changed){
		compute_encoder_value();
		display_encoder_value();
		send_encoder_data();
		encoder_changed = false;
	}

	/* Select the MIDI OUT stream (OUT from host) */
	Endpoint_SelectEndpoint(MIDI_STREAM_OUT_EPNUM);

	/* Check if endpoint is ready to be read from, if so discard its (unused) data */
	if (Endpoint_IsOUTReceived()){
		Endpoint_ClearOUT();
	}

}

// Send the midi data to the PC
void send_encoder_data(){
	while(!Endpoint_IsINReady()) {}	//wait for endpoint to be ready to be written *to*
	SendMIDICC(MIDI_CONTROLLER, encoder_value, 0, MIDI_CHANNEL);
}

// Get the encoder value on the display
void display_encoder_value(){
	DIGIT1_PORT = DIGIT_VALUES[encoder_value / 10];		//eek, built in knowledge that mask constrains us to 16 bits
	DIGIT2_PORT = DIGIT_VALUES[encoder_value % 10];		//eek, built in knowledge that mask constrains us to 16 bits
}

void compute_encoder_value(){
	uint8_t cur = ENCODER_SEQUENCE[encoder_state];
	uint8_t last = ENCODER_SEQUENCE[last_encoder_state];
	signed short inc = cur - last;
	if((last == 3) && (cur == 0)){	//forward rollover
		inc = 1;
	}
	else if((last == 0) && (cur == 3)){ //reverse rollover
		inc = -1;
	}
	encoder_value += inc;
	if(encoder_value < 0){
		encoder_value = 0;
	}
	if(encoder_value > 99){
		encoder_value = 99;
	}
}

void send_keypad_data(){
	
	for(uint8_t bit = 0; bit < 16; bit++){
		bool was_on = (last_keypad_state & (1 << bit));
		bool is_on = (keypad_state & (1 << bit));
		if(is_on != was_on){
			while(!Endpoint_IsINReady()) {}	//wait for endpoint to be ready to be written *to*
			SendMIDINoteChange(bit, is_on, 0, MIDI_CHANNEL);
			break;
		}
	}
}

EVENT_HANDLER(USB_Connect) {
    /* Start USB management task */
    Scheduler_SetTaskMode(USB_USBTask, TASK_RUN);
}

EVENT_HANDLER(USB_Disconnect) {
    /* Stop running audio and USB management tasks */
    Scheduler_SetTaskMode(USB_MIDI_Task, TASK_STOP);
    Scheduler_SetTaskMode(USB_USBTask, TASK_STOP);
}

EVENT_HANDLER(USB_ConfigurationChanged) {
    /* Setup MIDI stream endpoints */
    Endpoint_ConfigureEndpoint(MIDI_STREAM_OUT_EPNUM, EP_TYPE_BULK,
                               ENDPOINT_DIR_OUT, MIDI_STREAM_EPSIZE,
                               ENDPOINT_BANK_SINGLE);

    Endpoint_ConfigureEndpoint(MIDI_STREAM_IN_EPNUM, EP_TYPE_BULK,
                               ENDPOINT_DIR_IN, MIDI_STREAM_EPSIZE,
                               ENDPOINT_BANK_SINGLE);

    /* Start MIDI task */
    Scheduler_SetTaskMode(USB_MIDI_Task, TASK_RUN);
}
