static inline void Meters_Init() {
	// Timer2
	power_timer2_enable();

	// Timer 2 async setup (section 17.9)
	// a. Disable the Timer/Counter2 interrupts by clearing OCIE2x and TOIE2.
	TIMSK2 = 0;

	// b. Select clock source by setting AS2 as appropriate.
	//ASSR = _BV(AS2);

	// b(2). Wait for clock to come up.
	//_delay_ms(1000);

	// c. Write new values to TCNT2, OCR2x, and TCCR2x.
	TCNT2 = 0;
	OCR2A = 0;
	OCR2B = 0;

	TCCR2A = _BV(COM2A1) | _BV(COM2B1) | _BV(WGM21) | _BV(WGM20);
	TCCR2B = _BV(CS22) | _BV(CS21);

	// d. To switch to asynchronous operation: Wait for TCN2xUB, OCR2xUB, and TCR2xUB.
	//while( ASSR & (_BV(TCN2UB) | _BV(OCR2AUB) | _BV(OCR2BUB) | _BV(TCR2AUB) | _BV(TCR2BUB)) );

	// e. Clear the Timer/Counter2 Interrupt Flags.
	TIFR2 = 0;

	// f. Enable interrupts, if needed.
	TIMSK2 = _BV(TOIE2);

	// OC.2A
	DDRB |= 0b00010000;
	PORTB &= ~0b00010000;

	// OC.2B
	DDRD |= 0b00000010;
	PORTD &= ~0b00000010;
}

static inline void Switches_Init() {
	// Low four pins are the switch scan outputs.
	//		inactive: inputs pulled high (DDR=0,PORT=1,PUD=0)
	//		active: outputs pulled low (DDR=1,PORT=0)
	// High four pins are the switch scan inputs.
	//		always: inputs pulled high (DDR=0,PORT=1, PUD=0)

	DDRF &= ~0b11111111;
	PORTF |= 0b11111111;

	// Enable pull-up bus power by turning off PUD
	// (pull-up disable).
	MCUCR &= ~_BV(PUD);
}

const int debounce_counter_limit = 4;
static unsigned char key_debounce_counter[16];
static unsigned char key_active[16];

static void key_pressed(const unsigned char key) {
	static unsigned char message[] = { 0x90, 0x00 };
	message[1] = key;
	send_message(message, sizeof(message));
}

static void key_released(const unsigned char key) {
	static unsigned char message[] = { 0x80, 0x00 };
	message[1] = key;
	send_message(message, sizeof(message));
}

static void debounce_key(const unsigned char key, const unsigned char pressed) {
	if( key_active[key] ) {
		if( !pressed ) {
			key_active[key] = false;
			key_released(key);
		}
	} else {
		if( pressed ) {
			key_debounce_counter[key] += 1;
			if( key_debounce_counter[key] == debounce_counter_limit ) {
				key_active[key] = true;
				key_pressed(key);
			}
		} else {
			key_debounce_counter[key] = 0;
		}
	}
}

static void scan_buttons() {
	static unsigned char row = 0;

	unsigned char columns = PINF >> 4;
	unsigned char key = row * 4;

	debounce_key(key + 0, (columns & 1) == 0);
	debounce_key(key + 1, (columns & 2) == 0);
	debounce_key(key + 2, (columns & 4) == 0);
	debounce_key(key + 3, (columns & 8) == 0);

	// Set scan row pin output for next call -- allows
	// the signal to settle for a while.
	row = (row + 1) % 4;

	// Low four pins are the switch scan outputs.
	//		inactive: inputs pulled high (DDR=0,PORT=1,PUD=0)
	//		active: outputs pulled low (DDR=1,PORT=0)

	// Make pins inputs.
	// DDRF = 0b????0000 (PF[3:0] = input)
	DDRF &= ~0b00001111;

	// Change pin output values.
	// Make
	// row 0: PORT = PORT | 0x0F ^ 1, 0b????1110 (PF[3:1]   = pull-up, PF0 = Hi-Z)
	// row 1: PORT = PORT | 0x0F ^ 2, 0b????1101 (PF[3:2,0] = pull-up, PF1 = Hi-Z)
	// row 2: PORT = PORT | 0x0F ^ 4, 0b????1011 (PF[3,1:0] = pull-up, PF2 = Hi-Z)
	// row 3: PORT = PORT | 0x0F ^ 8, 0b????0111 (PF[  2:0] = pull-up, PF3 = Hi-Z)
	PORTF = (PORTF | 0x0F) ^ (1 << row);

	// Make one pin an output.
	// row 0: DDRF = 0b????0001 (PF[3:1  ] = input, PF0 = output low)
	// row 1: DDRF = 0b????0010 (PF[3:2,0] = input, PF1 = output low)
	// row 2: DDRF = 0b????0100 (PF[3,1:0] = input, PF2 = output low)
	// row 3: DDRF = 0b????1000 (PF[  2:0] = input, PF3 = output low)
	DDRF |= (1 << row);
}

ISR(TIMER2_OVF_vect) {
	scan_buttons();
}

/** Main program entry point. This routine configures the hardware required by the application, then
 *  starts the scheduler to run the application tasks.
 */
int xmain(void)
{
	/* Disable watchdog if enabled by bootloader/fuses */
	MCUSR &= ~(1 << WDRF);
	wdt_disable();

	/* Disable clock division */
	clock_prescale_set(clock_div_1);

	/* Hardware Initialization */
	Meters_Init();
	Switches_Init();

	/* Ring buffer Initialization */
	Buffer_Initialize(&Rx_Buffer);
	Buffer_Initialize(&Tx_Buffer);

	/* Indicate USB not ready */
	//UpdateStatus(Status_USBNotReady);

	/* Initialize Scheduler so that it can be used */
	Scheduler_Init();

	/* Initialize USB Subsystem */
	USB_Init();

	/* Scheduling - routine never returns, so put this last in the main function */
	Scheduler_Start();
}
